package malaria;

/**
 *
 * @author tra
 */

import java.util.*;

public class GenomicConfiguration {
    double s;
    double ts;
    int numDrugs;
    int[][] drugMuts;
    int numMuts;
    int totalClones;
    boolean isSI;
    boolean mutate;
    double probIntro;
    double precentReplaced;

    public GenomicConfiguration(double s, double ts, boolean isSI,
            int[][] drugMuts,
            boolean mutate, double probIntro, double percentReplaced) {
        this.s = s;
        this.ts = ts;
        this.isSI = isSI;
        this.drugMuts = drugMuts;
        this.numDrugs = drugMuts.length;
        int maxMut = 0;
        for (int[] dM : drugMuts) {
            for (int mut: dM) {
                if (maxMut<mut) maxMut = mut;
            }
        }
        this.numMuts = maxMut+1;
        this.totalClones = (int)Math.pow(2, maxMut+1);
        this.mutate = mutate;
        this.probIntro = probIntro;
        this.precentReplaced = percentReplaced;
    }
    
    public static GenomicConfiguration getStandard(
            double s, double ts, boolean isSI,
            int numDrugs, int mutsPerDrug,
            boolean mutate, double probIntro, double percentReplaced) {
        int[][] drugMuts = new int [numDrugs][mutsPerDrug];
        int currentMut = 0;
        for (int drug=0;drug<numDrugs; drug++) {
            for (int drugMut=0;drugMut<mutsPerDrug; drugMut++) {
                drugMuts[drug][drugMut] = currentMut;
                currentMut ++;
            }
        }
        return new GenomicConfiguration (s, ts, isSI, drugMuts,
                mutate, probIntro, percentReplaced);
    }

    /*
     * ACT case
     * mut 0 is shared by treatments (Artesunate)
     * each treatment has a "private" mutation
     */
    public static GenomicConfiguration getACT(
            double s, double ts, boolean isSI,
            int numDrugs,
            boolean mutate, double probIntro, double percentReplaced) {
        int[][] drugMuts = new int [numDrugs][2];
        for (int drug=0;drug<numDrugs; drug++) {
                drugMuts[drug][0] = 0;
                drugMuts[drug][1] = drug + 1;
        }
        return new GenomicConfiguration (s, ts, isSI, drugMuts,
                mutate, probIntro, percentReplaced);
    }


    public double[] getFreqsFixedForDrug(double drugResistance) {
        double[] freqs = new double[(int)Math.pow(2,drugMuts.length * drugMuts[0].length)];
        freqs[0] = 1.0 - drugResistance * drugMuts.length;
        for (int[] dM : drugMuts) {
            int pos = 0;
            for (int mut: dM) {
                if (mut>=0) {
                    pos += (int)Math.pow(2,mut);
                }
            }
            freqs[pos] = drugResistance;
        }
        return freqs;
    }

    public double[] getFreqsMultiForDrug(double drugResistance) {
        double[] freqs = new double[(int)Math.pow(2,drugMuts.length * drugMuts[0].length)];
        freqs[0] = 1.0 - drugResistance;
        freqs[freqs.length-1] = drugResistance;
        return freqs;
    }

    public double[] getFreqsLinkEqDrug(double totalSensitive) {
        double[] freqs = new double[(int)Math.pow(2, numMuts)];
        double[] partials = new double[numMuts];
        double perDrug = (1-totalSensitive);
        for(int[] muts : this.drugMuts) {
            for (int mut : muts) {
                if (mut>=0) {
                    partials[mut] = perDrug;
                }
            }
        }
        for (int i=0; i<freqs.length; i++) {
            freqs[i] = 1;
            for (int locus=0; locus<numMuts; locus++) {
                if ((i & (int)Math.pow(2, locus))>0) {
                    freqs[i] *= partials[locus];
                }
                else {
                    freqs[i] *= totalSensitive;
                }
            }
        }
        return freqs;
    }

    public static void main(String[] args) {
        int[][] drugMuts = {{0,1}, {2,3}};
        double[] freqs;
        double sum=0;
        GenomicConfiguration gc = new GenomicConfiguration(0.1, 0.01,
                false, drugMuts, false, 0, 0);
        freqs = gc.getFreqsLinkEqDrug(0.90);
        for (double freq: freqs) {
            sum += freq;
            System.out.print(freq + " ");
        }
        System.out.print("\n" + sum);
    }
}
