package malaria;

/**
 *
 * @author tra
 */

import java.util.*;

public class Recombination {
    int totalClones;
    double s;
    double ts;
    boolean isSI;
    ArrayList<int[]> tableRec;
    int[][] drugMuts;
    int epiType; //0 - full, 1 - duplication, 2 - mixed, 3 - "realistic"
    boolean mutate;
    double probIntro;
    double percentReplaced;
    Random generator;

    public Recombination(int totalClones, double s,
            double ts, boolean isSI, int[][] drugMuts, int epiType,
            boolean mutate, double probIntro, double percentReplaced) {
        this.totalClones = totalClones;
        this.s = s;
        this.ts = ts;
        this.isSI = isSI;
        this.drugMuts = drugMuts;
        this.epiType = epiType;
        tableRec = new ArrayList<int[]>();
        for(int i=0;i<totalClones*totalClones;i++) {
            tableRec.add(null); // Just to get size right
        }
        tableRecombination(); //table can be a verb...
        //System.out.println(mutate);
        this.mutate = mutate;
        this.probIntro = probIntro;
        this.percentReplaced = percentReplaced;
        generator = new Random();
    }

    public Recombination(GenomicConfiguration gc, int epiType) {
        this(gc.totalClones, gc.s, gc.ts, gc.isSI, gc.drugMuts, epiType,
                gc.mutate, gc.probIntro, gc.precentReplaced);
    }

    public int factorial(int n) {
        if (n<2) return 1;
        //hard coding for fast results
        if (n==2) return 2;
        if (n==3) return 6;
        if (n==4) return 24;
        if (n==5) return 120;
        if (n==6) return 720;
        int f=720;
        for (int i=7;i<=n;i++) f *= i;
        return f;

    }
    
    public double multinomial (int[] cases, double[] probs) {
        int nCases = 0;
        int denomPart = 1;
        for (int it : cases) {
            if (it==0) continue;
            nCases += it;
            denomPart *= factorial(it);
        }
        int numPart = factorial(nCases);
        double combin = numPart/denomPart;
        double prob = 1.0;
        for (int i=0; i<cases.length; i++) {
            //System.out.println(probs.get(i) + " " + cases.get(i) + " " + Math.pow(probs.get(i), cases.get(i)));
            prob *= Math.pow(probs[i], cases[i]);
        }
        //System.out.println("  p " + prob + " c " + combin + " nc " + nCases);
        return prob * combin;
    }
    
    public void tableRecombination() {
        for (int i1=0; i1<totalClones; i1++){
            for (int i2=0; i2<totalClones; i2++){
                int diffs = i1 ^ i2;
                int basis = (i1 | i2) ^ diffs;
                Vector<Integer> values = new Vector<Integer>();
                values.add(basis);
                //System.out.println(
                //        Integer.toBinaryString(i1)+"\t"+
                //        Integer.toBinaryString(i2)+"\t");
                int divDiffs = diffs;
                int divPos = 0;
                while (divDiffs >0) {
                    if (divDiffs % 2 == 1) {
                      Vector<Integer> dbl =  new Vector<Integer>(values);
                      for (int value : dbl) {
                          values.add(value + (int)Math.pow(2, divPos));
                      }
                    }
                    divPos += 1;
                    divDiffs /= 2;
                }
                int[] valuesArr = new int[values.size()];
                for (int i=0; i<valuesArr.length; i++) {
                    valuesArr[i] = values.get(i);
                }
                tableRec.set(i1*this.totalClones + i2, valuesArr);
            }
        }
    }

    public int calcNumMuts(int clone) {
        int numMuts = 0;
        while (clone>0) {
            numMuts += clone % 2;
            clone /= 2;
        }
        return numMuts;
    }
    
    public boolean hasAllMuts(int clone, int[] muts) {
        for (int mut: muts) {
            if ((clone & (int)Math.pow(2, mut)) == 0) return false;
        }
        return true;
    }

    public boolean hasOneMut(int clone, int[] muts) {
        for (int mut: muts) {
            if ((clone & (int)Math.pow(2, mut)) >0) return true;
        }
        return false;
    }

    /**
     * Check if the clone as the first mutation.
     * 
     * This a case where (either because of pk/pd or immunity) one
     * mutation is enough, but only the first one.
     * 
     * Think DHFR (as first) and DHPS (as second) for SP.
     * 
     * @param clone
     * @param muts
     * @return If the clone survives
     */
    public boolean hasFirstMut(int clone, int[] muts) {
        if ((clone & (int)Math.pow(2, muts[0])) >0) return true;
        return false;
    }



    public double[] calcRealS(int[] clones) {
        double[] res = new double[clones.length];
        double div = 0;
        for (int clone : clones) {
            div += Math.pow(1-s, calcNumMuts(clone));
        }
        for (int ci = 0;ci<res.length; ci++) {
            res[ci] = Math.pow(1-s,
                    calcNumMuts(clones[ci]))/div;
            //System.out.println(res[ci] + " " + calcNumMuts(clones[ci]));
        }
        //System.out.println();
        return res;
    }

    /* Recombination WITH fitness
     */
    public double[] recombine(int[] clones) {
        double[] clFreq = new double[totalClones];
        double[] clFit = calcRealS(clones);
        for (int c1=0; c1<clFit.length; c1++){
            for (int c2=0; c2<clFit.length; c2++){
                double sharedFit = clFit[c1] * clFit[c2];
                int[] newClones = tableRec.get(clones[c1]*totalClones + clones[c2]);
                for (int clone: newClones) {
                    clFreq[clone] += sharedFit / newClones.length;
                }
                //System.out.println(" " + totalClones +
                //        " " + c1 +  " " + c2 +  " " + sharedFit / newClones.length);
            }
        }
        return clFreq;
    }
    
    public double[][] calcTransmitingFreqs(int[] existingClones){
        double[][] f;
        
        if (epiType < 2) {
            f = new double[drugMuts.length +1][totalClones];
        }
        else {
            f = new double[(2*drugMuts.length) +1][totalClones];
        }

        f[0] = recombine(existingClones);
        for (int d=0;d<drugMuts.length;d++) {
            int[] drgMuts = drugMuts[d];
            Vector<Integer> survivingClones = new Vector<Integer>();
            Vector<Integer> survivingClonesDpl = new Vector<Integer>();
            for(int clone: existingClones) {
                if (epiType == 0) {
                    if (hasAllMuts(clone, drgMuts)) survivingClones.add(clone);
                }
                else if (epiType == 1) {
                    if (hasOneMut(clone, drgMuts)) survivingClones.add(clone);
                }
                else if (epiType == 2) {
                    if (hasAllMuts(clone, drgMuts)) survivingClones.add(clone);
                    if (hasOneMut(clone, drgMuts)) survivingClonesDpl.add(clone);
                }
                else {
                    if (hasAllMuts(clone, drgMuts)) survivingClones.add(clone);
                    if (hasFirstMut(clone, drgMuts)) survivingClonesDpl.add(clone);
                }
            }
            int[] svClArray =  new int[survivingClones.size()];
            for (int i=0;i<svClArray.length;i++) {
                svClArray[i] = survivingClones.get(i);
            }
            f[d+1] = recombine(svClArray);
            if (epiType > 1) {
                svClArray =  new int[survivingClonesDpl.size()];
                for (int i=0;i<svClArray.length;i++) {
                    svClArray[i] = survivingClonesDpl.get(i);
                }
                f[drugMuts.length+d+1] = recombine(svClArray);
            }
        }
        return f;
    }

    double calcMultiFreq(int[] clones, double[] cloneFreqs) {
        int[] cloneCounts = new int[totalClones];
        for (int clone : clones ) {
            //System.out.print(" "+clone);
            cloneCounts[clone] += 1;
        }
        //System.out.println();
        return multinomial(cloneCounts, cloneFreqs);
    }

    double[] introduceResistance(double[] freqs) {
        double numLoci = Math.log(this.totalClones)/Math.log(2);
        for (int locus=0; locus<numLoci; locus++) {
            if (generator.nextDouble() <= this.probIntro) {
                int mutAdd = (int)Math.pow(2, locus);
                /*
                //Introduction in all sensitives (divided)
                for (int clone=0;clone<this.totalClones; clone++) {
                    if ((clone & mutAdd) == 0) {
                        double qty = freqs[clone] * this.percentReplaced / (this.totalClones/2);
                        freqs[clone] -= qty;
                        freqs[clone + mutAdd] += qty;
                    }
                }
                */
                //Introduction in a single random clone
                int cloneMutate = generator.nextInt(this.totalClones);
                if ((cloneMutate & mutAdd) > 0) {
                    cloneMutate -= mutAdd;
                }
                double qty = freqs[cloneMutate] * this.percentReplaced;
                freqs[cloneMutate] -= qty;
                freqs[cloneMutate + mutAdd] += qty;
            }
        }
        return freqs;
    }

    double[] apply(double[] cloneFreqs, double[] drugFreqs,
            Map<Integer, Double> clonalMults,
            double[] residualFreqs) {
        double[] newFreqs = new double[totalClones];
        for (Map.Entry<Integer,Double> cm : clonalMults.entrySet()) {
            double sumMultiFreqs=0;
            int[] existingClones = new int[cm.getKey()];
            boolean done = false;
            while (!done) {
                double multiFreq = calcMultiFreq(existingClones, cloneFreqs);
                sumMultiFreqs += multiFreq;
                double[][] allFreqs = calcTransmitingFreqs(existingClones);
                for (int dCase = 0; dCase<= drugMuts.length; dCase++){
                    int siWeight = 1;
                    double weight;
                    if (dCase==0) {
                        weight = 1;
                        for (double dFreq: drugFreqs) {
                            weight -= dFreq;
                        }
                        weight *= cm.getValue();
                    }
                    else {
                        if (epiType > 1) {
                            weight = drugFreqs[dCase-1]/2 * cm.getValue();
                        }
                        else {
                            weight = drugFreqs[dCase-1] * cm.getValue();
                        }
                    }
                    if (isSI) {
                        siWeight = 0;
                        for (int clone : existingClones) {
                            if (allFreqs[dCase][clone] > 0) siWeight ++;
                        }
                    }
                    for (int ci=0;ci<totalClones;ci++) {
                        newFreqs[ci] += siWeight * multiFreq * weight * allFreqs[dCase][ci];
                        if (epiType>1 & dCase > 0) {
                            newFreqs[ci] += siWeight * multiFreq * weight * allFreqs[drugMuts.length + dCase][ci];
                        }
                    }
                }

                if (existingClones[0] == totalClones-1 && (
                        existingClones.length == 1 ||
                        existingClones[0] == existingClones[1]
                        )) {
                    done = true;
                }
                else {
                    for (int c=0;c<cm.getKey();c++) {
                        try {
                            if (existingClones[c] == totalClones ||
                                    existingClones[c] == existingClones[c+1]) { //combinations, not permutations
                                existingClones[c] = 0;
                            }
                            else {
                                existingClones[c] += 1;
                                break;
                            }
                        }
                        catch (ArrayIndexOutOfBoundsException e) {
                            existingClones[c] += 1;
                        }
                    }
                }
            }
        }
        /*
         * Calculate the effect of residual freqs
         * A negative value is a multiplier of drug freqs
         *   Say -0.25 makes reslevels = 0.25*dFreq
         */
        for (int i=0;i<newFreqs.length; i++) {
            for (int j=0;j<residualFreqs.length; j++) {
                if (!hasAllMuts(i, drugMuts[j])) {
                    //System.out.println("Recombination hpl " + i + " drug" + j);
                    if (residualFreqs[j]==1) { //this is stupid, but controls bugs
                        newFreqs[i] = 0;
                    }
                    else if (residualFreqs[j]>=0) {
                        newFreqs[i] *= (1 - residualFreqs[j]);
                    }
                    else {
                        newFreqs[i] *= (1 - (-residualFreqs[j]) * drugFreqs[j]);
                    }
                } 
            }
        }
        /* Calculate "transmission" */
        for (int i=0;i<newFreqs.length; i++) {
            newFreqs[i] *= Math.pow(1-ts, calcNumMuts(i));
        }
        /* We need to normalize (ie sum of freqs = 1)
         * This is because some partials will not some to 1
         *    (because of drug action killing sensitives)
         */
        double sumNewFreqs=0;
        for (double newFreq : newFreqs) sumNewFreqs += newFreq;
        for (int i=0;i<newFreqs.length; i++) {
            newFreqs[i] /= sumNewFreqs;
        }

        /* We now introduce resistant strains */
        if (this.mutate) {
            newFreqs = introduceResistance(newFreqs);
        }
        return newFreqs;
    }

    public static String freqsToString(double[] freqs) {
        String retStr = "";
        for (double freq : freqs) {
            retStr += Double.toString(freq) + " ";
        }
        return retStr;
    }
    
}
